/*
	Parse.cpp 
    Written by Saeed Amizadeh and Amir Nayyeri (PGen Project)

	This code is a parser for working with parse table resources
	generated by PGen ver 1.0 (.pt files),
	and can be compiled in any framework that supports ANSI C++ (including STL).

	Note: This code is not deterministic, and you can modify it as your needs 
		  but with your own responsibility. 
*/

#include "headers.h"
#include "scg.h"
#include "parse.h"



inline void CheckStream(istream& Stream) throw(CInvalidFormatException)
{
	if(!Stream.good())
	{
		Stream.clear();
		throw CInvalidFormatException("Stream is not in good condition");
	}
}

// CEdge Methods ////////////////////////////////////////////////////////////////////////

CEdge::CEdge(CNode* Head, CNode* Tail, bool isGraph): m_Head(Head), m_Tail(Tail),
m_IsGraph(isGraph) { }
//---------------------------------------------------------------------------------------

CNode* CEdge::GetHead() const { return m_Head; }
//---------------------------------------------------------------------------------------

CNode* CEdge::GetTail() const { return m_Tail; }
//---------------------------------------------------------------------------------------

STRING CEdge::GetInput() const { return m_Input; }
//---------------------------------------------------------------------------------------

STRING CEdge::GetAction() const { return m_Action; }
//---------------------------------------------------------------------------------------

bool CEdge::IsGraph() const { return m_IsGraph; }
//---------------------------------------------------------------------------------------

void CEdge::Load(istream& Stream) throw(CInvalidFormatException)
{
	char Length, Buffer[100];
	Stream.read((char*)&m_IsGraph,1);
	CheckStream(Stream);
	Stream.read((char*)&m_HeadID,4);
	CheckStream(Stream);

	Stream.read(&Length,1);
	CheckStream(Stream);
	Stream.read(Buffer,Length);
	CheckStream(Stream);
	Buffer[Length] = 0;
	m_Input = Buffer;
	
	Stream.read(&Length,1);
	CheckStream(Stream);
	if(Length>0)
	{
		Stream.read(Buffer,Length);
		CheckStream(Stream);
		Buffer[Length] = 0;
		m_Action = Buffer;
	}
}

// CNode Methods ////////////////////////////////////////////////////////////////////////

CNode::CNode(int ID): m_ID(ID), m_EdgeNum(0), m_IsFinal(false) {}
//---------------------------------------------------------------------------------------

CNode::~CNode()
{
	for(int i=0; i<m_EdgeNum; ++i) delete m_Edges[i];
	m_Edges.clear();
	m_EdgeNum = 0;
}
//---------------------------------------------------------------------------------------

int CNode::GetID() const { return m_ID; }
//---------------------------------------------------------------------------------------

bool CNode::IsFinal() const { return m_IsFinal; }
//---------------------------------------------------------------------------------------

CNode* CNode::JumpOverGraph(CCodeGenerator* CG) const
{
	for(int i=0; i<m_EdgeNum; ++i)
	{
		CEdge* p = m_Edges.at(i);
		if(p)
			if(p->IsGraph())
			{
				STRING Action = p->GetAction();
				if(Action.length() > 0) CG->Generate(Action, "", "");
				return p->GetHead();
			}
	}
	return NULL;
}
//---------------------------------------------------------------------------------------

bool CNode::NextState(const STRING& Token, CNode*& NextNode, STRING& Action) const
{
	for(int i=0; i<m_EdgeNum; ++i)
	{
		CEdge* p = m_Edges.at(i);
		if(p)
			if((!p->IsGraph()) && (p->GetInput() == Token))
			{
				Action = p->GetAction();
				NextNode = p->GetHead();
				return true;
			}
	}
	return false;
}
//---------------------------------------------------------------------------------------

bool CNode::NextGraph(CNode*& NextNode,const CParseTable* PT) const
{
	for(int i=0; i<m_EdgeNum; ++i)
	{
		CEdge* p = m_Edges.at(i);
		if(p)
			if(p->IsGraph())
			{
				NextNode = PT->GetFirstNodeOf(p->GetInput());
				return true;
			}
	}
	return false;
}
//---------------------------------------------------------------------------------------

void CNode::Load(istream& Stream) throw(CInvalidFormatException)
{
	Stream.read((char*)&m_ID,4);
	CheckStream(Stream);
	m_IsFinal = false;
	if(m_ID < 0)
	{
		m_ID *= -1;
		m_IsFinal = true;
	}

	Stream.read((char*)&m_EdgeNum,4);
	CheckStream(Stream);
	m_Edges.reserve(m_EdgeNum+2);
	for(int i=0; i<m_EdgeNum; ++i)
	{
		CEdge* p = new CEdge(NULL,NULL,false);
		p->Load(Stream);
		p->m_TailID = GetID();
		m_Edges.push_back(p);
	}
}

// CGraph Methods ///////////////////////////////////////////////////////////////////////

CGraph::CGraph(): m_NodeNum(0), m_Name("Untitled") { }
//---------------------------------------------------------------------------------------

CGraph::~CGraph()
{
	for(int i=0; i<m_NodeNum; ++i) delete m_Nodes[i];
	m_Nodes.clear();
	m_NodeNum = 0;
}
//---------------------------------------------------------------------------------------

void CGraph::Load(istream& Stream) throw(CInvalidFormatException)
{
	char Length, Buffer[100];
	Stream.read(&Length,1);
	CheckStream(Stream);
	Stream.read(Buffer,Length);
	CheckStream(Stream);
	Buffer[Length] = 0;
	m_Name = Buffer;
	Stream.read((char*)&m_NodeNum,4);
	CheckStream(Stream);

	m_Nodes.reserve(m_NodeNum+2);
	for(int i=0; i<m_NodeNum; ++i)
	{
		CNode* p = new CNode(i+1);
		p->Load(Stream);
		m_Nodes.push_back(p);
	}
	
	// Reconstructing pointers in memory:
	for(i=0; i<m_NodeNum; ++i)
	{
		CNode* p = m_Nodes.at(i);
		if(p)
			for(int j=0; j<p->m_EdgeNum; ++j)
			{
				CEdge* q = p->m_Edges.at(j);
				if(q)
					for(int k=0; k<m_NodeNum; ++k)
					{
						CNode* u = m_Nodes.at(k);
						if(u)
						{
							if(u->GetID() == q->m_HeadID) q->m_Head = u;
							if(u->GetID() == q->m_TailID) q->m_Tail = u;
						}
					}
			}
	}
}
//---------------------------------------------------------------------------------------

CNode* CGraph::GetFirstNode() const { return m_Nodes.at(0); }
//---------------------------------------------------------------------------------------

STRING CGraph::GetName() const { return m_Name; }

// CParseTable Methods //////////////////////////////////////////////////////////////////

CParseTable::CParseTable(): m_GraphNum(0) { }
//---------------------------------------------------------------------------------------

CParseTable::~CParseTable()
{
	for(int i=0; i<m_GraphNum; ++i) delete m_Graphs[i];
	m_Graphs.clear();
	m_GraphNum = 0;
}
//---------------------------------------------------------------------------------------

void CParseTable::Load(istream& Stream) throw(CInvalidFormatException)
{
	Stream.read((char*)&m_GraphNum,4);
	CheckStream(Stream);
	m_Graphs.reserve(m_GraphNum+2);

	for(int i=0; i<m_GraphNum; ++i)
	{
		CGraph* p = new CGraph;
		p->Load(Stream);
		m_Graphs.push_back(p);
	}
}
//---------------------------------------------------------------------------------------

CNode* CParseTable::GetStartNode() const { return m_Graphs[0]->GetFirstNode(); }
//---------------------------------------------------------------------------------------

CNode* CParseTable::GetFirstNodeOf(const STRING& Name) const
{
	for(int i=0; i<m_GraphNum; ++i)
	{
		CGraph* p = m_Graphs.at(i);
		if(p)
			if(p->GetName() == Name) return p->GetFirstNode();
	}
	return NULL;
}

// CParser Methods //////////////////////////////////////////////////////////////////////

CParser::CParser(istream& ParseFile, istream& in, ostream& out) throw(CInvalidFormatException)
{
	m_ParseTable = new CParseTable;
	m_ParseTable->Load(ParseFile);
	m_Scanner = new CScanner(in);
	m_CodeGenerator = new CCodeGenerator(out);
}
//---------------------------------------------------------------------------------------

CParser::~CParser()
{
	delete m_ParseTable;
	delete m_CodeGenerator;
	delete m_Scanner;
}
//---------------------------------------------------------------------------------------

void CParser::Parse() throw(CParseErrorException)
{
	CNODESTACK stack;
	BOOLSTACK  pps;
	STRING Token;
	bool AllowToReadNextToken = true, IsPostponed = false;

	CNode* temp = m_ParseTable->GetStartNode();
	if(!temp) return;

	stack.push(temp); pps.push(false);
	while(!stack.empty())
	{
		m_CurrentState = stack.top();
		IsPostponed = pps.top();
		stack.pop(); pps.pop();
		if (IsPostponed)
		{
			IsPostponed = false;
			m_CurrentState = m_CurrentState->JumpOverGraph(m_CodeGenerator);
		}

		if (AllowToReadNextToken)
		{
			Token = m_Scanner->NextToken();
			if((Token.length() <= 0)&&!m_CurrentState->IsFinal()) throw CParseErrorException("I Dont Know");/***********************/
		}

		STRING Action;
		if (m_CurrentState->NextState(Token, temp, Action) && (Token.length() > 0))
		{
			stack.push(temp); pps.push(false);
			if (Action.length() > 0) m_CodeGenerator->Generate(Action, Token, m_Scanner->lexeme());
			AllowToReadNextToken = true;
		}
		else if (m_CurrentState->NextGraph(temp, m_ParseTable))
		{
			AllowToReadNextToken = false;
			stack.push(m_CurrentState); pps.push(true);
			stack.push(temp);			pps.push(false);
		}
		else if (!m_CurrentState->IsFinal()) throw CParseErrorException("I Dont Know");/***********************/
		else AllowToReadNextToken = false;
	}
}
//---------------------------------------------------------------------------------------

CScanner* CParser::GetScanner() const { return m_Scanner; }
//---------------------------------------------------------------------------------------

CCodeGenerator* CParser::GetCodeGenerator() const { return m_CodeGenerator; }

